#include "inputclass.h"

InputClass::InputClass()
{
	controller_manager = NULL;
	touch_manager_ = NULL;
}


bool InputClass::Initialize(abfw::Platform& platform)
{
	controller_manager = platform.CreateSonyControllerInputManager();

	touch_manager_ = platform.CreateTouchInputManager();
	touch_manager_->EnablePanel(0);
	touch_manager_->EnablePanel(1);

	return true;
}

bool InputClass::Frame()
{
	for (int i = 0; i < 2; ++i)
	{
		for (int j = 0; j < 6; ++j)
		{
			previousTouchData[i][j].id = touch_manager_->GetTouch(i, j).id;
			previousTouchData[i][j].type = touch_manager_->GetTouch(i, j).type;
			previousTouchData[i][j].position = touch_manager_->GetTouch(i, j).position;
		}
	}

	controller_manager->Update();
	touch_manager_->Update();

	controller = controller_manager->GetController(0);

	//Handle analogue stick data
	leftXAxisVal = controller->left_stick_x_axis();
	leftYAxisVal = controller->left_stick_y_axis();

	rightXAxisVal = controller->right_stick_x_axis();
	rightYAxisVal = controller->right_stick_y_axis();

	if (leftXAxisVal != 0.0f || leftYAxisVal != 0.0f)
	{
		leftStickAngle = CalculateAngle(leftXAxisVal, leftYAxisVal);
		leftStickIntensity = CalculateIntensity(leftXAxisVal, leftYAxisVal);
	}

	if (rightXAxisVal != 0.0f || rightYAxisVal != 0.0f)
	{
		rightStickAngle = CalculateAngle(rightXAxisVal, rightYAxisVal);
		rightStickIntensity = CalculateIntensity(rightXAxisVal, rightYAxisVal);
	}

	return false;
}


void InputClass::Shutdown()
{
	if (controller_manager)
	{
		delete controller_manager;
		controller_manager = NULL;
	}

	if (touch_manager_)
	{
		delete touch_manager_;
		touch_manager_ = NULL;
	}

	return;
}

bool InputClass::AreButtonsPressed(UInt32 buttons)
{
	if (controller->buttons_down() & buttons == buttons)
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool InputClass::IsButtonPressed(UInt32 button)
{
	if (controller->buttons_down() & button)
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool InputClass::IsAnyButtonPressed()
{
	if (controller->buttons_down())
	{
		return true;
	}
	else
	{
		return false;
	}
}

UInt32 InputClass::ButtonPressed(UInt32 ignore_button)
{
	if (controller->buttons_down() & ignore_button)
	{
		return 0;
	}
	else
	{
		return controller->buttons_down();
	}
}

void InputClass::LeftStickMagnitude(float& x, float& y)
{
	x = leftXAxisVal;
	y = leftYAxisVal;
}

void InputClass::RightStickMagnitude(float& x, float& y)
{
	x = rightXAxisVal;
	y = rightYAxisVal;
}

void InputClass::ReturnLeftStickData(float& angle, float& intensity)
{
	angle = leftStickAngle;
	intensity = leftStickIntensity;
}

void InputClass::ReturnRightStickData(float& angle, float& intensity)
{
	angle = rightStickAngle;
	intensity = rightStickIntensity;
}

bool InputClass::IsTouchDown(int panel_num, int touch_num)
{
	if (touch_manager_->panel_enabled(panel_num))
	{
		const abfw::Touch& touch = touch_manager_->GetTouch(panel_num, touch_num);
		if (touch.type != abfw::TT_NONE)
		{
			return true;
		}
	}

	return false;
}

void InputClass::GetTouchPosition(int panel_num, int touch_num, float& posX, float& posY)
{
	if (touch_manager_->panel_enabled(panel_num))
	{
		const abfw::Touch& touch = touch_manager_->GetTouch(panel_num, touch_num);
		if (touch.type != abfw::TT_NONE)
		{
			posX = touch.position.x;
			posY = touch.position.y;
		}
	}
}

void InputClass::GetNewTouchPosition(int panel_num, int touch_num, float& posX, float& posY)
{
	if (touch_manager_->panel_enabled(panel_num))
	{
		const abfw::Touch& touch = touch_manager_->GetTouch(panel_num, touch_num);
		if ((previousTouchData[panel_num][touch_num].type == abfw::TT_NONE) && (touch.type != abfw::TT_NONE))
		{
			posX = touch.position.x;
			posY = touch.position.y;
		}
	}
}

bool InputClass::CalculateSwipe(int panel_num, int touch_num, float& swipeX, float& swipeY)
{
	if (touch_manager_->panel_enabled(panel_num))
	{
		return true;
	}

	return false;
}

float InputClass::CalculateIntensity(float x, float y)
{
	float intensity, pythagoras;

	pythagoras = ((powf(x, 2)) + (powf(y, 2)));

	intensity = sqrtf(pythagoras);

	return intensity;
}

float InputClass::CalculateAngle(float x, float y)
{
	float Angle;

	Angle = 0;

	if (x > 0.0f)
	{
		if (y < 0.0f)
		{
			Angle = 90.0f - (atan((fabsf(y) / x))* 180 / I_PI);
		}
		else if (y > 0.0f)
		{
			Angle = 180.0f - (atan((x / y)) * 180 / I_PI);
		}
		else
		{
			Angle = 90.0f;
		}
	}
	else if (x < 0.0f)
	{
		if (y < 0.0f)
		{
			Angle = 360.0f - (atan((fabsf(y) / fabsf(x))) * 180 / I_PI);
		}
		else if (y > 0.0f)
		{
			Angle = 270.0f - (atan((fabsf(x) / y)) * 180 / I_PI);
		}
		else
		{
			Angle = 270.0f;
		}
	}
	else
	{
		if (y < 0.0f)
		{
			Angle = 0.0f;
		}
		else if (y > 0.0f)
		{
			Angle = 180.0f;
		}
		else
		{
			Angle = 0.0f;
		}
	}

	return Angle;
}