#include "sceneclass.h"

SceneClass::SceneClass()
{
	input_ = NULL;
	platform_ = NULL;
	player_ptr_ = NULL;
	camera_position_ = b2Vec2(0.0f, 0.0f);
}

SceneClass::SceneClass(const SceneClass& other)
{
}

SceneClass::~SceneClass()
{
}

bool SceneClass::Initialize(Platform* platform, abfw::AudioManagerVita* audio_manager, InputClass* input, abfw::SpriteRenderer* renderer, abfw::Font& font, OptionsStorage* options_storage, const char* filename)
{
	abfw::PNGLoader png_loader;
	abfw::ImageData image_data;
	input_ = input;
	platform_ = platform;
	options_storage_ = options_storage;
	end_of_level_ = false;
	timer_ = 0.0f;

	audio_manager_ = audio_manager;

	background_image_ = new AnimatedSprite;

	background_image_->LoadTexture("background.png", *platform);
	background_image_->set_width(1024);
	background_image_->set_height(1024);
	background_image_->set_position(abfw::Vector3(platform->width() / 2, platform->height() / 2, 0.9f));

	m_objectStore.Initialize();

	// create the Box2D physics world
	b2Vec2 gravity(0.0f, -10.0f);
	physics_world_ = new b2World(gravity);

	LoadSceneFile(renderer, font, filename);

	UIObject* uibar = new UIValueBar;

	static_cast<UIValueBar*>(uibar)->Initialize(*platform_, "bar.png", LEFT, 100, 36, 20, player_ptr_->GetHealthPointer(), player_ptr_->GetMaxHealthPointer());

	ui_objects_.push_back(uibar);

	uibar = NULL;

	uibar = new UIValueBar;

	static_cast<UIValueBar*>(uibar)->Initialize(*platform_, "bar.png", BOTTOM, 450, platform_->height() - 36, 52, player_ptr_->GetFuelPointer(), player_ptr_->GetMaxFuelPointer());

	ui_objects_.push_back(uibar);

	return true;
}

void SceneClass::Shutdown()
{
	while (game_objects_.size() > 0)
	{
		game_objects_.back()->Shutdown();
		delete game_objects_.back();
		game_objects_.pop_back();
	}

	while (ui_objects_.size() > 0)
	{
		delete ui_objects_.back();
		ui_objects_.pop_back();
	}

	if (physics_world_)
	{
		delete physics_world_;
		physics_world_ = NULL;
	}
	
}

int SceneClass::Frame(float dt)
{
	current_dt_ = dt;
	
	if (!end_of_level_)
	{
		if (input_->IsButtonPressed(ABFW_SONY_CTRL_START))
			return 2;

		timer_ += dt;

		physics_world_->Step(dt, 1, 1);

		float new_uv_y = background_image_->uv_position().y - (dt / 2.0f);

		if (new_uv_y >= 1.0f)
			new_uv_y -= 1.0f;

		background_image_->set_uv_position(abfw::Vector2(0.0f, new_uv_y));

		if (end_timer_ > 0.0f)
		{
			end_timer_ -= dt;
			if (end_timer_ <= 0.0f)
				end_of_level_ = true;
		}

		if (player_ptr_)
		{
			camera_position_ = player_ptr_->GetWorldPosition();
		}
		
		//process game objects
		for (list<GameObject*>::iterator object = game_objects_.begin(); object != game_objects_.end(); object++)
		{
			ObjectFrame(*object);
		}

		//process ui objects
		for (list<UIObject*>::iterator object = ui_objects_.begin(); object != ui_objects_.end(); object++)
		{
			(*object)->Frame(dt);
		}

		//do generic frame processing for objects
		for (list<GameObject*>::iterator object = game_objects_.begin(); object != game_objects_.end(); object++)
		{
			(*object)->FrameGeneric(camera_position_, dt);
		}
		
		//clean up the necessary parts of the scene
		for (list<GameObject*>::iterator object = delete_list_.begin(); object != delete_list_.end(); object++)
		{
			DeleteObject(*object);
		}

		delete_list_.clear();
	}
	else
	{
		if (player_ptr_)
		{
			return 4;
		}
		else
		{
			return 5;
		}
		
	}

	return 1;
}

void SceneClass::ObjectFrame(GameObject* current_object)
{
	//determine the object type and do its unique frame calculations
	if (strcmp(current_object->typeTag_, "PlayerObject") == 0)
	{
		dynamic_cast<PlayerObject*>(current_object)->Frame(input_, current_dt_);
		return;
	}

	if (strcmp(current_object->typeTag_, "AreaHazard") == 0)
	{
		dynamic_cast<AreaHazard*>(current_object)->Frame(current_dt_);
		return;
	}

	if (strcmp(current_object->typeTag_, "ProjectileObject") == 0)
	{
		dynamic_cast<ProjectileObject*>(current_object)->Frame(current_dt_);
		return;
	}

	if (strcmp(current_object->typeTag_, "EnemyObject") == 0)
	{
		if (player_ptr_)
		{
			dynamic_cast<EnemyObject*>(current_object)->Frame(player_ptr_->GetWorldPosition(), current_dt_);
		}
		else
		{
			dynamic_cast<EnemyObject*>(current_object)->Frame(b2Vec2(0.0f, 0.0f), current_dt_);
		}

		return;
	}

	if (strcmp(current_object->typeTag_, "AnimationObject") == 0)
	{
		dynamic_cast<AnimationObject*>(current_object)->Frame(current_dt_);
		return;
	}
}

void SceneClass::DeleteObject(GameObject* object)
{
	if (object)
	{
		if (strcmp(object->typeTag_, "PlayerObject") == 0)
		{
			end_timer_ = player_ptr_->GetDeathAnimationTime();
			player_ptr_ = NULL;
		}
		object->Shutdown();
		game_objects_.remove(object);
		if (object->GetObjectBody())
			std::cout << "Deleting " << object->typeTag_ << std::endl;
			physics_world_->DestroyBody(object->GetObjectBody());
		delete object;
	}
}

GameObject* SceneClass::InstantiateObject(GameObject* object, float posX, float posY, float rot)
{
	GameObject* new_object = NULL;

	if (strcmp(object->typeTag_, "ProjectileObject") == 0)
	{
		new_object = new ProjectileObject;
		static_cast<ProjectileObject*>(new_object)->InstantiateCopy(*(ProjectileObject*)object, posX, posY, rot);
		game_objects_.push_back(new_object);
	}

	if (strcmp(object->typeTag_, "AnimationObject") == 0)
	{
		new_object = new AnimationObject;
		static_cast<AnimationObject*>(new_object)->InstantiateCopy(*(AnimationObject*)object, posX, posY, rot);
		game_objects_.push_back(new_object);
	}
	
	return new_object;
}

void SceneClass::Render(abfw::SpriteRenderer* renderer_, abfw::Font& font)
{
	if (background_image_)
		background_image_->Render(renderer_, font);

	for (list<GameObject*>::iterator object = game_objects_.begin(); object != game_objects_.end(); object++)
	{
		if ((*object)->visible_)
			(*object)->Render(renderer_, font);
	}

	for (list<UIObject*>::iterator object = ui_objects_.begin(); object != ui_objects_.end(); object++)
	{
		if ((*object)->visible_)
			(*object)->Render(renderer_, font);
	}

	font.RenderText(renderer_, abfw::Vector3(950.0f, 10.0f, -0.9f), 1.0f, 0xff009999, abfw::TJ_RIGHT, "%d Enemies Remaining", enemy_count_);
}

void SceneClass::LoadSceneFile(abfw::SpriteRenderer* renderer, abfw::Font& font, const char* filename)
{
	int num_objects;
	int i = 0;

	int bufferPosition = 0;
	char input;
	bool result;

	File* level_file = platform_->CreateFile();
	File* object_File = platform_->CreateFile();
	
	UIValueBar* loading_bar = new UIValueBar;
	loading_bar->Initialize(*platform_, "bar.png", LEFT, 640, 160, 320, &i, &num_objects);

	bool success = level_file->Open(filename);
	if (success)
	{
		char* buffer = NULL;
		char* objectBuffer = NULL;
		char* objFilename = NULL;
		char* objType = NULL;

		Int32 file_size;
		Int32 obj_size;
		success = level_file->GetSize(file_size);
		if (success)
		{
			buffer = new char[file_size];
			Int32 bytes_read;
			success = level_file->Read(buffer, file_size, bytes_read);

			if (success)
				level_file->Close();

			if (success)
			{
				ParseInt(buffer, bufferPosition, num_objects);
				bufferPosition++;

				//Read in all object data for the level
				for (i = 0; i < num_objects; i++)
				{
					//Read up to the data
					ParseToData(buffer, bufferPosition);
					input = buffer[bufferPosition];

					//Read in the object filename
					ParseCharArray(buffer, bufferPosition, objFilename);
					bufferPosition++;

					float new_uv_x = background_image_->uv_position().x + (1.0f/(float)num_objects);

					if (new_uv_x >= 1.0f)
						new_uv_x -= 1.0f;

					background_image_->set_uv_position(abfw::Vector2(new_uv_x, 0.0f));
					
					//Render the current loading progress
					platform_->PreRender();
					renderer->Begin();
					background_image_->Render(renderer, font);
					font.RenderText(renderer, abfw::Vector3(480.0f, 240.0f, -0.9f), 1.0f, 0xff00ffff, abfw::TJ_CENTRE, "Loading File: %s", objFilename);
					loading_bar->Frame(0.0f);
					loading_bar->Render(renderer, font);
					renderer->End();
					platform_->PostRender();

					//Open and read the correct object file
					success = object_File->Open(objFilename);
					if (success)
					{
						success = object_File->GetSize(obj_size);
						if (success)
						{
							objectBuffer = new char[obj_size];
							Int32 read_bytes;
							success = object_File->Read(objectBuffer, obj_size, read_bytes);

							if (success)
								object_File->Close();

							if (success)
							{
								int objectBufferPosition = 0;

								//Read in the object type
								ParseCharArray(objectBuffer, objectBufferPosition, objType);
								objectBufferPosition++;

								//Read in the object position here and move the object to that position
								float posX, posY, rot;

								ParseFloat(buffer, bufferPosition, posX);
								bufferPosition += 1;

								ParseFloat(buffer, bufferPosition, posY);
								bufferPosition += 1;

								ParseFloat(buffer, bufferPosition, rot);
								bufferPosition += 2;

								rot = rot * (b2_pi / 180.0f);

								//Assign the correct derived type to the current game object
								GameObject* next_object;
								result = m_objectStore.SetObjectType(next_object, objType);
								next_object->owner_scene_ = this;
								if (result)
								{
									next_object->LoadDataFile(objFilename, platform_, physics_world_, audio_manager_, input_, options_storage_, posX, posY, rot);
								}

								if (strcmp(next_object->typeTag_, "PlayerObject") == 0 && !player_ptr_)
								{
									player_ptr_ = (PlayerObject*)next_object;
								}

								if (strcmp(next_object->typeTag_, "EnemyObject") == 0)
								{
									enemy_count_++;
								}

								game_objects_.push_back(next_object);

								delete[] objectBuffer;
								objectBuffer = NULL;

								

							}
						}
					}
				}
			}
		}
	}
}

void SceneClass::ResetLevel()
{
	timer_ = 0.0f;
}