#include "sprite_application.h"
#include <graphics/sprite_renderer.h>
#include <system/platform.h>
#include <assets/png_loader.h>
#include <graphics/image_data.h>
#include <graphics/texture.h>
#include <iostream>
#include "box2d_helpers.h"


SpriteApplication::SpriteApplication(abfw::Platform& platform) :
	abfw::Application(platform),
	sprite_renderer_(NULL)
{
}

SpriteApplication::~SpriteApplication()
{
}

void SpriteApplication::Init()
{
	// load the font to draw the on-screen text
	bool font_loaded = font_.Load("comic_sans", platform_);
	if(!font_loaded)
	{
		std::cout << "Font failed to load." << std::endl;
		exit(-1);
	}

	input_ = new InputClass;

	input_->Initialize(platform_);

	audio_manager_ = new abfw::AudioManagerVita;

	// create a sprite renderer to draw the sprites
	sprite_renderer_ = platform_.CreateSpriteRenderer();

	options_storage_.Initialize(audio_manager_);

	title_scene_.Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, NULL);

	pause_menu_.Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, NULL);

	options_menu_.Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, NULL);

	win_scene_.Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, "You Win!");
	win_scene_.SetReturnValue((int)GAME_WIN);

	lose_scene_.Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, "You Lose!");
	lose_scene_.SetReturnValue((int)GAME_LOSE);

	scene_ = NULL;

	frame_time_ = 0.0f;
}

void SpriteApplication::CleanUp()
{
	delete input_;
	input_ = NULL;

	// free up the sprite renderer
	delete sprite_renderer_;
	sprite_renderer_ = NULL;

	delete audio_manager_; 
	audio_manager_ = NULL;

}

bool SpriteApplication::Update(float ticks)
{
	frame_time_ += ticks;

	// calculate the frame rate
	frame_rate_ = 1.0f / ticks;

	input_->Frame();
	
	//update the correct game state
	switch (game_state_)
	{
	case TITLE_SCREEN:
		game_state_ = (GameState)title_scene_.Frame(ticks);
		if (game_state_ == OPTIONS_MENU)
			options_menu_.returnValue_ = 0;
		break;
	case IN_GAME:
		game_state_ = (GameState)scene_->Frame(ticks);
		break;
	case OPTIONS_MENU:
		game_state_ = (GameState)options_menu_.Frame(ticks);
		break;
	case GAME_WIN:
		game_state_ = (GameState)win_scene_.Frame(ticks);
		if (game_state_ == TITLE_SCREEN)
		{
			scene_->Shutdown();
			delete scene_;
			scene_ = NULL;
		}
		break;
	case GAME_LOSE:
		game_state_ = (GameState)lose_scene_.Frame(ticks);
		if (game_state_ == TITLE_SCREEN)
		{
			scene_->Shutdown();
			delete scene_;
			scene_ = NULL;
		}
		break;
	case PAUSE_MENU:
		game_state_ = (GameState)pause_menu_.Frame(ticks);
		if (game_state_ == TITLE_SCREEN)
		{
			scene_->Shutdown();
			delete scene_;
			scene_ = NULL;
		}
		if (game_state_ == OPTIONS_MENU)
			options_menu_.returnValue_ = 2;
		break;
	
	}

	if (!scene_ && game_state_ == IN_GAME)
	{
		scene_ = new SceneClass;

		scene_->Initialize(&platform_, audio_manager_, input_, sprite_renderer_, font_, &options_storage_, options_storage_.GetLevelName(options_storage_.GetCurrentLevel()));
	}

	return true;
}

void SpriteApplication::Render()
{
	// set up sprite renderer for drawing
	sprite_renderer_->Begin();

	//
	//render the correct game state
	switch (game_state_)
	{
	case TITLE_SCREEN:
		title_scene_.Render(sprite_renderer_, font_);
		break;
	case IN_GAME:
		scene_->Render(sprite_renderer_, font_);
		break;
	case PAUSE_MENU:
		pause_menu_.Render(sprite_renderer_, font_);
		break;
	case OPTIONS_MENU:
		options_menu_.Render(sprite_renderer_, font_);
		break;
	case GAME_WIN:
		win_scene_.Render(sprite_renderer_, font_);
		break;
	case GAME_LOSE:
		lose_scene_.Render(sprite_renderer_, font_);
		break;
	}

	font_.RenderText(sprite_renderer_, abfw::Vector3(850.0f, 510.0f, -0.9f), 1.0f, 0xff009999, abfw::TJ_LEFT, "FPS: %.1f", frame_rate_);

	// tell sprite renderer that all sprites have been drawn
	sprite_renderer_->End();
}